#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <strstream.h>
#include <math.h>
#include <unistd.h>
#include <GL/gl.h>
#include "glut.h"
#include "Text.h"
#include "action.h"
#include "endovar.h"
#include "relmech.h"

int actionCount = 0; Action *actions[10];
int endovarCount = 0; Endovar *endovars[10];
int relmechCount = 0; Relmech *relmechs[10];

int activeActionIndex = -1;
double maxActionValue = 0;

void init(void);
void loop(void);
void displayValue(void);
void displayFatigue(void);
void displayEndo(void);
void keyboard(unsigned char, int, int);

void addAction(const Text &);
Action *findAction(const Text &);

void addEndovar(const Text &);
Endovar *findEndovar(const Text &);

void addRelmech(const Text &);
Relmech *findRelmech(const Text &);

long valueID, endoID, fatigueID;

static int                processScript(const Text &);
static int                processLine(istrstream &);

// -------------------------- command functions --------------------------

static int actionCmd(istrstream &); 
static int colorActionCmd(istrstream &); 
static int colorEndovarCmd(istrstream &); 
static int endovarCmd(istrstream &); 
static int inhibitCmd(istrstream &); 
static int relmechCmd(istrstream &); 
static int bindActionToEndovarCmd(istrstream &); 
static int bindEndovarToActionCmd(istrstream &); 
static int bindRelmechToEndovarCmd(istrstream &); 
static int blockActionByRelmechCmd(istrstream &); 

// -------------------------- main ---------------------------------------

void main(int argc, char *argv[])
{

  if (argc != 2) {
    cerr << argc << endl;
    cerr << "Usage: actsel scriptName\n";
    exit(-1);
  }

  if (!processScript(argv[1])) exit(-1);

  init();
  glutIdleFunc(loop); 
  glutMainLoop();
}

void loop(void)
{
  // update endogenous variables
  for (int i=0;i<endovarCount;i++) endovars[i]->updateValue();

  // update fatigues before computing the value for the actions
  for (i=0;i<actionCount;i++) actions[i]->updateFatigue();
  // update the action values
  for (i=0;i<actionCount;i++) actions[i]->updateValue();

  // check which action is active
  maxActionValue = -1; 
  for (i=0;i<actionCount;i++) {
    double value = actions[i]->getValue();
    if (value > maxActionValue) { 
      maxActionValue = value; activeActionIndex = i;
    }
  }
  
  displayValue(); displayFatigue(); displayEndo(); 
  // schedule the next callback 
  glutIdleFunc(loop); 
  sginap(10);
}

int
processScript(const Text &name)
{

  // process the script
  fstream        scriptFile;
  Text scriptFilename = name;
  scriptFile.open(scriptFilename.get(), ios::in);
  if (!scriptFile) {
    cerr << "can not open: " << scriptFilename << endl;
    return 0;
  }

  char *line = new char [256];
  int lineCount = 1;
  while(scriptFile.getline(line,256,'\n')) {
    istrstream iss(line, scriptFile.gcount());
    if (!processLine(iss)) {
      cerr << endl;
      cerr << "*******************************************************\n";
      cerr << "error at line: " << lineCount << " of '";
      cerr << scriptFilename << "': " << line << endl;
      cerr << "*******************************************************\n";
      return 0;
    }
    lineCount++;
  }
  delete [] line;

  scriptFile.close();
  return 1;
}

int
processLine(istrstream &iss)
{
  Text word;  iss >> word;

  if (!word || word == "//" || word == "quit") return 1;
  else if (word == "action") { return actionCmd(iss); } 
  else if (word == "colorAction") { return colorActionCmd(iss); } 
  else if (word == "colorEndovar") { return colorEndovarCmd(iss); } 
  else if (word == "bindActionToEndovar") 
    { return bindActionToEndovarCmd(iss); } 
  else if (word == "bindEndovarToAction") 
    { return bindEndovarToActionCmd(iss); } 
  else if (word == "bindRelmechToEndovar") 
    { return bindRelmechToEndovarCmd(iss); } 
  else if (word == "blockActionByRelmech") 
    { return blockActionByRelmechCmd(iss); } 
  else if (word == "endovar") { return endovarCmd(iss); } 
  else if (word == "inhibit") { return inhibitCmd(iss); } 
  else if (word == "relmech") { return relmechCmd(iss); } 

  cerr << "unrecognized command: " << "___" << word << "___" << endl;
  return 0;
}

// ------------------------- command functions ------------------------------

int actionCmd(istrstream &iss) 
{
  Text actionName;
  double fatigueGain, fatigueLoss;
  iss >> actionName >> fatigueGain >> fatigueLoss;
  
  // add a new action
  addAction(actionName);

  // verify that the new action has been added successfully
  Action *action = findAction(actionName); assert(action);

  // set the fatigue parameters
  action->setFatigueGain(fatigueGain);
  action->setFatigueLoss(fatigueLoss);
  return 1;
}

int colorActionCmd(istrstream &iss) 
{
  Text actionName; int r, g, b;
  iss >> actionName >> r >> g >> b;
  
  // find the right action
  Action *action = findAction(actionName); 
  if (!action) {
    cerr << "colorAction -- could not find action: ";
    cerr << actionName << endl;
    return 0;
  }

  action->setColor((GLubyte) r,(GLubyte) g,(GLubyte) b);

  return 1;
}

int colorEndovarCmd(istrstream &iss) 
{
  Text endovarName; int r, g, b;
  iss >> endovarName >> r >> g >> b;
  
  // find the right endovar
  Endovar *endovar = findEndovar(endovarName); 
  if (!endovar) {
    cerr << "colorEndovar -- could not find endovar: ";
    cerr << endovarName << endl;
    return 0;
  }

  endovar->setColor((GLubyte) r,(GLubyte) g,(GLubyte) b);

  return 1;
}

int bindActionToEndovarCmd(istrstream &iss) 
{
  Text endovarName, actionName;
  double weight;
  iss >> actionName >> endovarName >> weight;
  
  // find the right action
  Action *action = findAction(actionName); 
  if (!action) {
    cerr << "bindActionToEndovar -- could not find action: ";
    cerr << actionName << endl;
    return 0;
  }

  // find the right endogenous variable
  Endovar *endovar = findEndovar(endovarName); 
  if (!endovar) {
    cerr << "bindActionToEndovar -- could not find endogenous variable: ";
    cerr << endovarName << endl;
    return 0;
  }

  // bind the variable to the action
  endovar->addAction(action, weight);

  return 1;
}

int bindEndovarToActionCmd(istrstream &iss) 
{
  Text endovarName, actionName;
  double weight;
  iss >> endovarName >> actionName >> weight;
  
  // find the right action
  Action *action = findAction(actionName); 
  if (!action) {
    cerr << "bindEndovarToAction -- could not find action: ";
    cerr << actionName << endl;
    return 0;
  }

  // find the right endogenous variable
  Endovar *endovar = findEndovar(endovarName); 
  if (!endovar) {
    cerr << "bindEndovarToAction -- could not find endogenous variable: ";
    cerr << endovarName << endl;
    return 0;
  }

  // bind the variable to the action
  action->addEndovar(endovar, weight);

  return 1;
}

int bindRelmechToEndovarCmd(istrstream &iss) 
{
  Text endovarName, relmechName;
  double activeGain, inactiveGain;
  iss >> relmechName >> endovarName >> activeGain >> inactiveGain;
  
  // find the right relmech
  Relmech *relmech = findRelmech(relmechName); 
  if (!relmech) {
    cerr << "bindRelmechToEndovar -- could not find releasing mechanism: ";
    cerr << relmechName << endl;
    return 0;
  }

  // find the right endogenous variable
  Endovar *endovar = findEndovar(endovarName); 
  if (!endovar) {
    cerr << "bindRelmechToEndovar -- could not find endogenous variable: ";
    cerr << endovarName << endl;
    return 0;
  }

  // bind the variable to the releasing mechanism
  endovar->addRelmech(relmech, activeGain, inactiveGain);

  return 1;
}

int blockActionByRelmechCmd(istrstream &iss) 
{
  Text actionName, relmechName, status; int flag;
  iss >> actionName >> relmechName >> status;
  
  // find the right action
  Action *action = findAction(actionName); 
  if (!action) {
    cerr << "blockActionByRelmech -- could not find action: ";
    cerr << actionName << endl;
    return 0;
  }

  // find the right relmech
  Relmech *relmech = findRelmech(relmechName); 
  if (!relmech) {
    cerr << "blockActionByRelmech -- could not find releasing mechanism: ";
    cerr << relmechName << endl;
    return 0;
  }

  if (status == "ON" || status == "on") flag = 1; else flag = 0;

  // bind the variable to the releasing mechanism
  action->addBlock(relmech, flag);

  return 1;
}

int endovarCmd(istrstream &iss) 
{
  Text endovarName;
  double autoGain, initValue, minValue;
  iss >> endovarName >> autoGain >> initValue >> minValue;
  
  // add a new endogenous variable
  addEndovar(endovarName);

  // verify that the new endogenous variable has been added successfully
  Endovar *endovar = findEndovar(endovarName); assert(endovar);

  // set the fatigue parameters
  endovar->setAutoGain(autoGain);
  endovar->setValue(initValue);
  endovar->setMinValue(minValue);

  return 1;
}

int inhibitCmd(istrstream &iss) 
{
  Text inhibitorName, inhibiteeName;
  double value;
  iss >> inhibitorName >> inhibiteeName >> value;
  
  // find the inhibitor
  Action *inhibitor = findAction(inhibitorName); 
  if (!inhibitor) {
    cerr << "inhibitCmd -- could not find action: ";
    cerr << inhibitorName << endl;
    return 0;
  }

  // find the inhibitor
  Action *inhibitee = findAction(inhibiteeName); 
  if (!inhibitee) {
    cerr << "inhibitCmd -- could not find action: ";
    cerr << inhibiteeName << endl;
    return 0;
  }

  inhibitee->addInhibitor(inhibitor,2.0);

  return 1;
}

int relmechCmd(istrstream &iss) 
{
  Text relmechName, activity;
  iss >> relmechName >> activity;
  
  // add a new releasing mechanism
  addRelmech(relmechName);

  // verify that the new releasing mechanism has been added successfully
  Relmech *relmech = findRelmech(relmechName); assert(relmech);

  if (activity == "ON" || activity == "on") relmech->toggle();

  return 1;
}

// ------------------------- misc functions ---------------------------------

void addAction(const Text &name)
{
  assert(actionCount < 10);
  actions[actionCount] = new Action; assert(actions[actionCount]);
  actions[actionCount]->setName(name);
  actionCount++;
}

Action *
findAction(const Text &name)
{

  for (int i=0;i<actionCount;i++) {
    Action *action = actions[i];
    if (action->getName() == name) return action;
  }
  
  return NULL;
}
 
void addEndovar(const Text &name)
{
  assert(endovarCount < 10);
  endovars[endovarCount] = new Endovar; assert(endovars[endovarCount]);
  endovars[endovarCount]->setName(name);
  endovarCount++;
}

Endovar *
findEndovar(const Text &name)
{

  for (int i=0;i<endovarCount;i++) {
    Endovar *endovar = endovars[i];
    if (endovar->getName() == name) return endovar;
  }
  
  return NULL;
}

void addRelmech(const Text &name)
{
  assert(relmechCount < 10);
  relmechs[relmechCount] = new Relmech; assert(relmechs[relmechCount]);
  relmechs[relmechCount]->setName(name);
  relmechCount++;
}

Relmech *
findRelmech(const Text &name)
{

  for (int i=0;i<relmechCount;i++) {
    Relmech *relmech = relmechs[i];
    if (relmech->getName() == name) return relmech;
  }
  
  return NULL;
}

// --------------------------- graphical display -----------------------------

void init(void)
{

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowPosition(800, 200);
  glutInitWindowSize(450,100);
  valueID = glutCreateWindow("value");
  glutDisplayFunc(displayValue);
  glutKeyboardFunc(keyboard);

  // initialize the projection 
  glMatrixMode(GL_PROJECTION);
  glOrtho(0,100,0,11.0, -1, 1);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();

  glutInitWindowPosition(800, 340);
  endoID = glutCreateWindow("endo");
  glutDisplayFunc(displayEndo);
  // initialize the projection 
  glMatrixMode(GL_PROJECTION);
  glOrtho(0,100,0,1.1, -1, 1);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();

  glutInitWindowPosition(800, 480);
  fatigueID = glutCreateWindow("fatigue");
  glutDisplayFunc(displayFatigue);
  // initialize the projection 
  glMatrixMode(GL_PROJECTION);
  glOrtho(0,100,0,1.1, -1, 1);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClearDepth(1.0);
  glutSwapBuffers();

}

void displayValue(void)
{
  // update the value window
  glutSetWindow(valueID);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // display the name of the current activity
  assert(activeActionIndex >= 0 && activeActionIndex < actionCount);
  Action *activeAction = actions[activeActionIndex];
  Text &activeName = activeAction->getName();
  glColor3ub(200,200,200);
  glRasterPos2f(80.9,8.95);
  for (int i=0; i<activeName.getLen();i++) {
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, activeName[i]);
  }

  // display present releasing mechainisms
  for (int j=0;j<relmechCount;j++) {
    Relmech *relmech = relmechs[j];
    if (relmech->isActive()) {
      Text &relmechName = relmech->getName();
      glColor3ub(200,200,200);
      glRasterPos2f(10.0,8.95-1.5*j);
      for (int i=0; i<relmechName.getLen();i++) {
	glutBitmapCharacter(GLUT_BITMAP_8_BY_13, relmechName[i]);
      }
    }
  }
  // draw the action values
  for (i=0;i<actionCount;i++) actions[i]->drawValue();
  glutSwapBuffers();

}

void displayFatigue(void)
{
  // update the value window
  glutSetWindow(fatigueID);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  for (int i=0;i<actionCount;i++) actions[i]->drawFatigue();
  glutSwapBuffers();

}

void displayEndo(void)
{
  // update the value window
  glutSetWindow(endoID);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  for (int i=0;i<endovarCount;i++) endovars[i]->drawGL();
  glutSwapBuffers();

}

void keyboard(unsigned char key, int, int)
{
  switch (key) {
  case 3: // ctrl-c
    exit(1);
    break;
  case 27: // esc
    exit(1);
    break;
  case 101: // e
    Relmech *enemy = findRelmech("enemy");
    if (enemy) enemy->toggle();
    break;
  case 119: // w
    Relmech *water = findRelmech("water");
    if (water) water->toggle();
    break;
  case 102: // f
    Relmech *food = findRelmech("food");
    if (food) food->toggle();
    break;
  case 116: // t
    Relmech *toy = findRelmech("toy");
    if (toy) toy->toggle();
    break;
  default: 
    cerr << "unspecified key pressed: " << (int) key << endl;
  };

}
